#include <stdio.h>
#include <conio.h>
#include <process.h>
#include <values.h>
#include <math.h>

#define MAX 52
#undef STDVYSTUP

/* struktura reprezentuje jeden vrchol, ukladaji se do ni vsechny potrebne
	 informace o vrcholu */
typedef struct {
	float vzdalenost;			// nastavi se na zacatku
	float cena_benzinu;   // nastavi se na zacatku
	float celkova_cena;   // meni se behem vypoctu
	int otevreny;					/* meni se behem vypoctu (vrchol uzavru pote, co na
													 nem provedu vypocet - projdu jeho nasledniky */
	/* mohla by tu byt promenna int predchudce, kam by se ukladal vrchol, ze
	ktereho se sem doslo, diky tomu by bylo mozne vypsat jednotlive body
	nejlevnejsi cesty a ne jenom finalni cenu */
} t_vrchol;

/* globalni promenne */
float nadrz_max;		  // objem nadrze vozu
float mil_na_galon;   // kolik mili ujede vuz na jeden galon
int index_posledni;   // index do pole vrcholu ukazujici na cilove mesto
											// pouzivam ho jako maximalni index pri vypoctech

/* funkce pouzivana pri ladeni, vypise obsah pole vrcholu */
void vypis( t_vrchol *v ) {
	int i;
	printf("\ncislo uzlu | vzdalenost | cena_benzinu | celkova_cena\n");
	for (i=0; i<=index_posledni; i++ )
		printf("%2d: %f %f %f\n", i, v[i].vzdalenost, v[i].cena_benzinu, v[i].celkova_cena );
	printf("\n");
}

/* funkce zapise cenu nejlevnejsi trasy do vystupniho souboru*/
void zapis_vysledek( FILE *fw, t_vrchol *v, int dataset ) {
	float vysledek;
	vysledek = v[index_posledni].celkova_cena / 100;
	#ifdef STDVYSTUP
		fw = stdout;
	#endif
	fprintf( fw, "Data Set #%d\n", dataset );
	fprintf( fw, "minimum cost = $%.2f\n", vysledek );
}

/* najde index v poli vrcholu, kde je vrchol s minimalni celkovou cenou,
	 vsimam si pouze vrcholu, ktere jsem jeste neuzavrel */
int najdi_index_min( t_vrchol *v ) {
	int i, index=1;
	float min=MAXFLOAT;
	for (i=1; i<index_posledni; i++) {
		if ( (v[i].otevreny == 1) && (v[i].celkova_cena < min) ) {
			index = i;
			min = v[i].celkova_cena;
		}
	}
	/* jen pro jistotku, v normalnim pripade nemuze nastat */
	if ( min == MAXFLOAT ) {
		printf("ERROR - MAXFLOAT");
		exit(1);
	}
	/* */
	return index;
}

/* provede zaokrouhleni floatu na cele cislo
	 - je potreba, protoze kazda platba u benzinky se zaokrouhluje na nejblizsi
		 cent */
float zaokrouhli( float x ) {
	if ( (ceil(x)-x) < (x-floor(x)) )
		return ceil(x);
	else return floor(x);
}

/* **************************************************** */
/* funkce provadejici samotny vypocet nejlevnejsi trasy */
void spocitej( t_vrchol *v ) {
	int aktualni = 0;		/* index do pole vrcholu ukazujici na prave zpracovavany
												 uzel */
	float na_plnou = nadrz_max * mil_na_galon; /* jakou vzdalenost auto
																								ujede na plnou nadrz */
	int nalezeno_pulka; // nasel jsem benzinku dal nez na pul nadrze (true==1)
	int nalezeno; // nasel jsem nejakou benzinku v dosahu (true==1);
	int i;
	float pulka, dojezd, cena;

	while ( aktualni < index_posledni ) {
		nalezeno_pulka = 0;
		nalezeno = 0;
		i = aktualni + 1;
		pulka = v[aktualni].vzdalenost + na_plnou/2;
		dojezd = v[aktualni].vzdalenost + na_plnou;
		/* nejdrive jen prolitnu vrcholy, ktere jsou bliz nez na pul nadrze, kdyz
			 narazim na cil, tak koncim, jelikoz posledni cesta uz me nic nestoji -
			 nasel jsem tedy nejlevnejsi cestu */
		while ( v[i].vzdalenost <= pulka ) {
			if ( i == index_posledni ) {
				v[i].celkova_cena = v[aktualni].celkova_cena;
				return;
			}
			i++;
		}
		/* prochazim vrcholy, ktere jsou dal nez pulka, ale stale jsou v dosahu
			 na jednu nadrz;
			 kdyz najdu cil, opet hned koncim */
		while ( v[i].vzdalenost <= dojezd ) {
			if ( i == index_posledni ) {
				v[i].celkova_cena = v[aktualni].celkova_cena;
				return;
			}
			nalezeno_pulka = nalezeno = 1;
			cena = zaokrouhli(v[aktualni].celkova_cena + v[i].cena_benzinu * (v[i].vzdalenost-v[aktualni].vzdalenost) / mil_na_galon + 200);
			if ( cena < v[i].celkova_cena ) {
				v[i].celkova_cena = cena;
			}
			i++;
		}

		/* pokud se mi nepodarilo najit zadnou benzinku, do ktere jsem schopen
			 dojet a ktera je dal nez co ujedu na pul nadrze, musim zkusit benzinky,
			 ktere jsou bliz */
		if ( nalezeno_pulka == 0 ) {
			i = aktualni+1;
			while ( v[i].vzdalenost <= pulka ) {
				nalezeno = 1;
				cena = zaokrouhli(v[aktualni].celkova_cena + v[i].cena_benzinu * (v[i].vzdalenost-v[aktualni].vzdalenost) / mil_na_galon + 200);
				if ( cena < v[i].celkova_cena ) {
					v[i].celkova_cena = cena;
				}
				i++;
			}
		}

		if ( nalezeno == 0 ) {
			printf("CHYBA - CESTA NEEXISTUJE\n");
			exit(1);
		}
		v[aktualni].otevreny = 0;
		aktualni = najdi_index_min(v);
	}
	/* jen pro jistotku, v normalnim pripade se sem nikdy nelze dostat */
	printf("ERROR - while\n");
	exit(1);
}

/* ******************************************************* */
/* main funkce
	 provadi postupne nacitani ze vstupniho souboru, zapis ziskanych informaci
	 do pole vrcholu, pote zavola vypocet a pote zapsani vysledku */
int main (int argc, char* argv[]) {
	FILE *fr, *fw;
	t_vrchol vrcholy[MAX];
	float vzdalenost_cile, cena_na_startu;
	int i, pocet_benzinek, dataset=0;

	// kontrola spravneho spusteni a pristupnosti souboru
	if ( argc < 3 ) {
		printf( "CHYBA - SPATNE PARAMETRY\nspravny zpusob: budgettr.exe vstupni_soubor vystupni_soubor" );
		return(1);
	}
	printf( "Vstupni soubor: %s\nVystupni soubor: %s\n", argv[1], argv[2] );
	fr = fopen( argv[1], "rt" );
	if ( fr == NULL ) {
		printf( "CHYBA - NELZE OTEVRIT VSTUPNI SOUBOR\n" );
		return(1);
	}
	if ( (fw = fopen( argv[2], "wt")) == NULL ) {
		printf( "CHYBA - NELZE OTEVRIT VYSTUPNI SOUBOR\n" );
		fclose(fr);
		return(1);
	}

	/* ************************************************** */
	while (1) {
		dataset++;  // urcuje kolikatou vstupni sadu zpracovavam
		fscanf( fr, "%f", &vzdalenost_cile );
		// zaporne cislo indikuje konec vstupniho souboru
		if ( vzdalenost_cile < 0 )
			break;
		fscanf( fr, "%f %f %f %d", &nadrz_max, &mil_na_galon, &cena_na_startu, &pocet_benzinek );
		if ( pocet_benzinek > MAX-2 ) {
			printf("CHYBA - MAXIMALNI POCET BENZINEK JE %d\n", MAX);
			fclose(fr); fclose(fw);
			return(1);
		}
		// prvni vrchol je start, uz v nem jsem zaplatil za celou nadrz, cena benzinu je irelevantni
		vrcholy[0].vzdalenost = 0;
		vrcholy[0].cena_benzinu = 0;
		vrcholy[0].celkova_cena = cena_na_startu * 100; //cena_na_startu je v dolarech, vse ostatni je v centech
		for (i=1; i<=pocet_benzinek; i++) {
			fscanf( fr, "%f %f", &vrcholy[i].vzdalenost, &vrcholy[i].cena_benzinu );
			vrcholy[i].celkova_cena = MAXFLOAT;
			vrcholy[i].otevreny = 1;
			if ( vrcholy[i].vzdalenost > vzdalenost_cile ) {
				printf("CHYBA - BENZINKA NEMUZE BYT DALE NEZ JE CIL\n");
				fclose(fr); fclose(fw);
				return(1);
			}
		}
		index_posledni =  pocet_benzinek + 1;
		// posledni vrchol je cil, v nem uz benzin nekupuji, proto je cena benzinu nulova
		vrcholy[index_posledni].vzdalenost = vzdalenost_cile;
		vrcholy[index_posledni].cena_benzinu = 0;
		vrcholy[index_posledni].celkova_cena = MAXFLOAT;

		spocitej( &vrcholy[0] );
		zapis_vysledek( fw, &vrcholy[0], dataset );
	}
	/* ************************************************** */

	// jen uklidim a konec
	fclose(fr);
	fclose(fw);
	printf( "Vypocet proveden.\n" );
	return(0);
}

/*
	Co by se dalo vylepsit:
	- kontrolovat spravnost vstupniho souboru (pripustne hodnoty, spravny pocet
		benzinek apod.)
	- vypsat nejlevnejsi cestu
	- moznost vypisu na stdout
	- moznost zadani nazvu vstupniho a vystupniho souboru
	- lepe osetrit situaci, kdy cesta neexistuje
	- vrcholy nemit v pevnem poli, ale dynamicky alokovat
*/